/* ``The contents of this file are subject to the Erlang Public License,
 * Version 1.0, (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.erlang.org/EPL1_0.txt
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is Erlang-4.7.3, December, 1998.
 * 
 * The Initial Developer of the Original Code is Ericsson Telecom
 * AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
 * Telecom AB. All Rights Reserved.
 * 
 * Contributor(s): ______________________________________.''
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <string.h>

#include "ei.h"
#include "putget.h"

int ei_decode_string(const char *buf, int *index, char *p)
{
  const char *s = buf + *index;
  const char *s0 = s;
  int len;
  int i;
  int etype;

  switch (get8(s)) {
  case ERL_STRING_EXT:
    len = get16be(s);

    if (p) {
      memmove(p,s,len); 
      p[len] = (char)0;
    }
    s += len;
    break;

  case ERL_LIST_EXT:
    /* Really long strings are represented as lists of small integers.
     * We don't know in advance if the whole list is small integers,
     * but we decode as much as we can, exiting early if we run into a
     * non-character in the list.
     */
    len = get32be(s);
    for (i=0; i<len; i++) {
      if ((etype = get8(s)) != ERL_SMALL_INTEGER_EXT) return -1;
      if (p) p[i] = get8(s);
      else s++;
    }
    break;

  case ERL_NIL_EXT:
    if (p) p[0] = (char)0;
    break;
    
  default:
    return -1;
  }

  *index += s-s0; 

  return 0; 
}
